#!/usr/bin/env python3
"""
Parse multiple sources and output SWF format or an adapted SWF format called BSCSWF. Supported input formats are:

 - sacct output
 - BSC's Simulator [4] [5] sim_mgr (unprocessed) trace file
 - Standard Workflow Format [1] [2]
 - Walfredo Cirne's model output [3]

[1] Chapin, S.J. et al. (1999). Benchmarks and Standards for the Evaluation of Parallel Job Schedulers.
In: Feitelson, D.G., Rudolph, L. (eds) Job Scheduling Strategies for Parallel Processing. JSSPP 1999.
Lecture Notes in Computer Science, vol 1659. Springer, Berlin, Heidelberg.

[2] Standard Workflow Format specification can be found at https://www.cs.huji.ac.il/labs/parallel/workload/swf.html.

[3] Cirne, Walfredo, and Francine Berman. "A comprehensive model of the supercomputer workload." Proceedings of the
Fourth Annual IEEE International Workshop on Workload Characterization. WWC-4 (Cat. No. 01EX538). IEEE, 2001.

[4] Jokanovic, Ana, Marco D'Amico, and Julita Corbalan. "Evaluating slurm simulator with real-machine slurm and vice
versa." 2018 IEEE/ACM Performance Modeling, Benchmarking and Simulation of High Performance Computer Systems (PMBS).
IEEE, 2018.

[5] CES BSC adapted version of BSC's SLURM Simulator can be found at
https://earth.bsc.es/gitlab/mgimenez/ces_slurm_simulator.

BSC-CNS - Earth Sciences
Manuel Giménez de Castro Marciani
2023
"""

import pandas as pd
import datetime
import locale  # this is to read date files


def strfseconds(input: str):
    """
    Reads a string formatted as [D-]HH:MM:SS and returns the time in seconds, if possible. Otherwise -1.
    :param input: formatted string
    :return: either the total in seconds or -1
    """
    days = 0
    try:
        split_hyphen = input.split("-")
    except AttributeError:
        # this is not a string, we return
        return -1
    if len(split_hyphen) == 2:
        try:
            days = int(split_hyphen[0])
        except ValueError:
            # the format is not the expected one if we cannot cast the left part of the hyphen to an integer
            return -1
        time = split_hyphen[1]
    else:
        time = split_hyphen[0]

    split_dots = time.split(":")

    if len(split_dots) != 3:
        # if there are not 3 values if we split at :, then we are not in the expected format
        return -1

    try:
        hours = int(split_dots[0])
        minutes = int(split_dots[1])
        seconds = int(split_dots[2])
    except ValueError:
        # if we fail to cast to integer any of the split values, then we are out of the expected format
        return -1

    return days * 3600 * 24 + hours * 3600 + minutes * 60 + seconds


def strfkbs(input: str):
    """
    reads a string as formatted by acct and returns an integer representing it in kilobytes
    :param input: string as formated by sacct with memory usage
    :return: memory in kilobytes
    """
    if "G" in input:
        try:
            gigabytes = int(input.strip("G"))
        except ValueError:
            # couldn't cast, then there is some gibberish in the field
            return -1
        # 1024*1024*gigabytes
        return 1048576 * gigabytes

    if "M" in input:
        try:
            megabytes = int(input.strip("M"))
        except ValueError:
            # couldn't cast, then there is some gibberish in the field
            return -1
        return 1024 * megabytes

    if "K" in input:
        try:
            kilobytes = int(input.strip("K"))
        except ValueError:
            # couldn't cast, then there is some gibberish in the field
            return -1
        return kilobytes

    # If it doesn't enter any of the previous cases then it is 0n, which is the null usage recording of memory, or gibberish
    return -1


def strfdatelinux(value) -> datetime.datetime:
    """
    This function parses a date as given by the output of "date" program in Linux.
    :param value: string with the date
    :return: datetime object with the parsed date
    """
    locale.setlocale(locale.LC_ALL, locale.normalize('en_US.utf8'))  # set to US standard 
    try:
        return datetime.datetime.strptime(value, "%a %b %d %H:%M:%S %Z %Y")
    except ValueError:
        print(f"[WARNING] {value} failed to create valid datetime object for start time. Ignoring timezone name.")
        split_value = value.split()
        if len(split_value) == 6:
            value = " ".join(split_value[:4] + [split_value[5]])  # remove timezone name
            try:
                return datetime.datetime.strptime(value, "%a %b %d %H:%M:%S %Y")
            except ValueError:
                print(
                    f"[WARNING] {value} is not in the expected date format without timezone (e.g. Mon Nov 17 14:01:47 2008). Ignoring it.")
        else:
            print(
                f"[WARNING] {value} is not in the expected date format for start time (e.g. Mon Nov 17 14:01:47 CET 2008). Ignoring it.")


class SWFtrace:
    def __init__(self, swf_version, computer_name, location, acknowledge, conversion, max_jobs, max_records, preemption,
                 unix_start_time, time_zone_string, start_time, end_time, max_nodes, max_procs, max_runtime, max_memory,
                 allow_overuse, max_queues, queues, queue, max_partitions, partitions, partition, note, job_list,
                 file_location):
        # HEADER PARSER
        self._swf_version = swf_version
        self._computer = computer_name
        self._location = location
        self._acknowledge = acknowledge
        self._conversion = conversion
        self._max_jobs = max_jobs
        self._max_records = max_records
        self._preemption = preemption
        self._unix_start_time = unix_start_time
        self._time_zone_string = time_zone_string
        self._start_time = start_time
        self._end_time = end_time
        self._max_nodes = max_nodes
        self._max_procs = max_procs
        self._max_runtime = max_runtime
        self._max_memory = max_memory
        self._allow_overuse = allow_overuse
        self._max_queues = max_queues
        self._queues = queues
        self._queue = queue
        self._max_partitions = max_partitions
        self._partitions = partitions
        self._partition = partition
        self._note = note

        # pandas dataframe
        self._job_list = job_list

        self._input_file = file_location

    @classmethod
    def from_swf(cls, input_file):
        swf_version = "2.2"
        computer_name = ""
        location = ""
        acknowledge = ""
        conversion = ""
        max_jobs = ""
        max_records = ""
        preemption = ""
        unix_start_time = ""
        time_zone_string = ""
        start_time = ""
        end_time = ""
        max_nodes = ""
        max_procs = ""
        max_runtime = ""
        max_memory = ""
        allow_overuse = ""
        max_queues = ""
        queues = ""
        queue = []
        max_partitions = ""
        partitions = ""
        partition = []
        note = ""

        # parse header file
        with open(input_file, "r") as file:
            for line_num, line in enumerate(file):
                if line[0] == ";":
                    # this is a comment line. Then we remove the first character.
                    # TODO there is a more elegant way to do this in python.

                    try:
                        field, value = line[1:].rstrip("\n").split(": ")
                    except ValueError:
                        # this is an empty comment line, so we do nothing
                        continue

                    # remove white spaces
                    field = field.lstrip().rstrip()
                    value = value.rstrip().lstrip()
                    if field == "Version":
                        swf_version = value
                    elif field == "Computer":
                        computer_name = value
                    elif field == "Installation":
                        location = value
                    elif field == "Information":
                        information = value
                    elif field == "Acknowledge":
                        acknowledge = value
                    elif field == "Conversion":
                        # TODO HOW TO DEAL WITH MULTI LINES
                        conversion = value
                    elif field == "UnixStartTime":
                        try:
                            unix_start_time = int(value)
                        except ValueError:
                            print(f"[WARNING] Failed to cast to int {field} with value '{value}' in line {line_num}. "
                                  f"Ignoring it.")
                    elif field == "TimeZone":
                        # this field is ignored in current versions of SWF
                        pass
                    elif field == "TimeZoneString":
                        time_zone_string = value
                    elif field == "StartTime":
                        start_time = strfdatelinux(value)
                    elif field == "EndTime":
                        end_time = strfdatelinux(value)
                    elif field == "MaxJobs":
                        try:
                            max_jobs = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "MaxRecords":
                        try:
                            max_records = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "MaxNodes":
                        try:
                            max_nodes = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "MaxProcs":
                        try:
                            max_procs = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "Preemption":
                        if value.casefold() in ["yes", "no", "double", "ts"]:
                            preemption = value
                        else:
                            print(f"[WARNING] Value '{value}' in line {line_num} for Preemption field is incorrect. "
                                  f"Ignoring it")
                    elif field == "Queues":
                        queues = value
                    elif field == "Queue":
                        # TODO HOW TO DEAL WITH MULTI LINES
                        pass
                    elif field == "MaxQueues":
                        try:
                            max_queues = int(value)
                        except ValueError:
                            print(
                                f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}. "
                                f"Ignoring it.")
                    elif field == "Partitions":
                        partitions = value
                    elif field == "Partition":
                        # TODO HOW TO DEAL WITH MULTI LINES
                        pass
                    elif field == "MaxPartitions":
                        try:
                            max_partitions = int(value)
                        except ValueError:
                            print(f"[WARNING] Failed to cast to int at {field} with value '{value}' in line {line_num}."
                                  f"Ignoring it.")
                    elif field == "Note":
                        note = value
                    else:
                        print(f"[WARNING] Unrecognized field {field} in the header in line {line_num}. Ignoring it.")
                else:
                    # we stop reading the file since we are not out of the header
                    break

        job_list = pd.read_csv(input_file, names=["job_id", "submit_time", "wait_time", "runtime", "alloc_cpus",
                                                  "avg_cpu_used", "used_memory", "requested_cpus", "requested_time",
                                                  "requested_memory", "status", "uid", "gid", "app_id", "qos_id",
                                                  "partition_id", "preceding", "thinktime"],
                               sep="\s+", comment=";")

        return cls(swf_version, computer_name, location, acknowledge, conversion, max_jobs, max_records, preemption,
                   unix_start_time, time_zone_string, start_time, end_time, max_nodes, max_procs, max_runtime,
                   max_memory, allow_overuse, max_queues, queues, queue, max_partitions, partitions, partition, note,
                   job_list, input_file)

    @classmethod
    def from_saact(cls, input_file, fwf: bool = False, anon: bool = False):
        """
        Converts a sacct output to SWF,

        :param input_file: input sacct output file
        :param fwf: boolean flag indicating if the workload is fixed width or not
        :param anon: boolean flag to control if anonymize uid and gid
        :return:
        """
        swf_version = "2.2"
        computer_name = ""
        location = ""
        acknowledge = ""
        conversion = ""
        max_jobs = ""
        max_records = ""
        preemption = ""
        unix_start_time = ""
        time_zone_string = ""
        max_nodes = ""
        max_procs = ""
        max_runtime = ""
        max_memory = ""
        allow_overuse = ""
        max_queues = ""
        queue = []
        max_partitions = ""
        partition = []
        note = ""

        if fwf:
            sacct_trace = pd.read_fwf(input_file, infer_nrows=2)
        else:
            sacct_trace = pd.read_csv(input_file, sep="|")

        # FILTERING
        # get rid of the line with hyphens.
        # checking if the line is COMPLETE out of hyphens, otherwise do nothing
        if fwf:
            appended_values = "".join(sacct_trace.iloc[0])
            if len(appended_values) == appended_values.count("-"):
                sacct_trace.drop(0, inplace=True)

        # get rid of steps. If the jobid has a dot, then we know that it is part of the original's id
        sacct_trace = sacct_trace[~sacct_trace["JobID"].str.contains(".", regex=False)]

        # convert to numeric cpu count fields
        sacct_trace[["AllocCPUS", "ReqCPUS"]] = sacct_trace[["AllocCPUS", "ReqCPUS"]].apply(pd.to_numeric,
                                                                                            downcast='integer')

        if not sacct_trace["AveCPU"].isna().all():
            sacct_trace["AveCPU"] = pd.to_datetime(sacct_trace["AveCPU"], format='%H:%M:%S', errors='coerce')
            # sacct_trace[["AveCPU"]] = sacct_trace[["AveCPU"]].apply(pd.to_datetime)

        # convert to date the appropriate fields: submit, waittime, elapsed, timelimit
        sacct_trace["Submit"] = pd.to_datetime(sacct_trace["Submit"], format="%Y-%m-%dT%H:%M:%S", errors='coerce')
        sacct_trace["Start"] = pd.to_datetime(sacct_trace["Start"], format="%Y-%m-%dT%H:%M:%S", errors='coerce')

        # create a new column with the wait time
        sacct_trace["Waittime"] = sacct_trace['Start'] - sacct_trace['Submit']

        # convert timedate to time delta
        for i, row in sacct_trace.iterrows():
            elapsed_seconds = strfseconds(row["Elapsed"])
            timelimit_seconds = strfseconds(row["Timelimit"])
            if elapsed_seconds != -1:
                sacct_trace.at[i, "Elapsed"] = datetime.timedelta(seconds=elapsed_seconds)
            if timelimit_seconds != -1:
                sacct_trace.at[i, "Timelimit"] = datetime.timedelta(seconds=timelimit_seconds)
            else:
                sacct_trace.at[i, "Timelimit"] = -1

        # keep track of the start and finish of the simulation's jobs
        start_time = sacct_trace['Submit'].min()
        end_time = (sacct_trace['Submit'] + sacct_trace['Elapsed']).max()
        for i, row in sacct_trace.iterrows():
            # finally, transform submit to elapsed seconds since the start of the simulation
            sacct_trace.at[i, "Submit"] = (sacct_trace.at[i, "Submit"] - start_time).total_seconds()

        # Memory convert
        for i, row in sacct_trace.iterrows():
            sacct_trace.at[i, "ReqMem"] = strfkbs(row["ReqMem"])

        # SWF status codes
        for i, row in sacct_trace.iterrows():
            if "BOOT_FAIL" in row["State"]:
                sacct_trace.at[i, "State"] = 0
            elif "CANCELLED" in row["State"]:
                sacct_trace.at[i, "State"] = 5
            elif "COMPLETED" in row["State"]:
                sacct_trace.at[i, "State"] = 1
            elif "DEADLINE" in row["State"]:
                sacct_trace.at[i, "State"] = 0
            elif "FAILED" in row["State"]:
                sacct_trace.at[i, "State"] = 0
            elif "NODE_FAIL" in row["State"]:
                sacct_trace.at[i, "State"] = 0
            elif "OUT_OF_MEMORY" in row["State"]:
                sacct_trace.at[i, "State"] = 0
            elif "PENDING" in row["State"]:
                sacct_trace.at[i, "State"] = -1
            elif "PREEMPTED" in row["State"]:
                sacct_trace.at[i, "State"] = 0
            elif "RUNNING" in row["State"]:
                sacct_trace.at[i, "State"] = -1
            elif "REQUEUED" in row["State"]:
                sacct_trace.at[i, "State"] = -1
            elif "RESIZING" in row["State"]:
                sacct_trace.at[i, "State"] = -1
            elif "REVOKED" in row["State"]:
                sacct_trace.at[i, "State"] = -1
            elif "SUSPENDED" in row["State"]:
                sacct_trace.at[i, "State"] = -1
            elif "TIMEOUT" in row["State"]:
                sacct_trace.at[i, "State"] = 0
            else:  # invalid code
                sacct_trace.at[i, "State"] = -1

        if anon:
            # we factorize to anonymize data. We get rid of the index table
            sacct_trace["JobID"] = pd.factorize(sacct_trace["JobID"])[0]
            sacct_trace["UID"] = pd.factorize(sacct_trace["UID"])[0]
            sacct_trace["GID"] = pd.factorize(sacct_trace["GID"])[0]

        # queues and partitions we factorize, but keep the dictionary since they carry hard restrictions on the usage
        sacct_trace["QOS"], queues = pd.factorize(sacct_trace["QOS"])
        sacct_trace["Partition"], partitions = pd.factorize(sacct_trace["Partition"])

        # cast, again, dates to integer. Its final form
        for i, row in sacct_trace.iterrows():
            sacct_trace.at[i, "Waittime"] = row["Waittime"].total_seconds()
            sacct_trace.at[i, "Elapsed"] = row["Elapsed"].total_seconds()
            if row["Timelimit"] != -1:
                sacct_trace.at[i, "Timelimit"] = row["Timelimit"].total_seconds()

        sacct_trace[["Waittime", "Elapsed", "Timelimit"]] = sacct_trace[["Waittime", "Elapsed", "Timelimit"]].apply(
            pd.to_numeric, downcast='integer', errors='coerce')

        # finally, set to -1 all empty fields
        sacct_trace.fillna(-1, inplace=True)

        # TODO: this is the bare minimum of fields for swf. We can include an extra column "backfilled" that will no be
        # used at the swf
        job_list = sacct_trace[["JobID", "Submit", "Waittime", "Elapsed", "AllocCPUS", "AveCPU", "ReqCPUS", "Timelimit",
                                "State", "UID", "GID", "QOS", "Partition"]]

        job_list.rename({"JobID": "job_id", "Submit": "submit_time", "Waittime": "wait_time", "Elapsed": "runtime",
                         "AveCPU": "avg_cpu_used", "AllocCPUS": "alloc_cpus", "ReqCPUS": "requested_cpus",
                         "Timelimit": "requested_time", "State": "status", "UID": "uid", "GID": "gid",
                         "QOS": "qos_id", "Partition": "partition_id"}, axis='columns', inplace=True)

        # set to -1 all of those fields not provided by saact
        job_list["used_memory"] = -1
        # job_list["requested_memory"] = -1
        job_list["app_id"] = -1
        job_list["preceding"] = -1
        job_list["thinktime"] = -1

        return cls(swf_version, computer_name, location, acknowledge, conversion, max_jobs, max_records, preemption,
                   unix_start_time, time_zone_string, start_time, end_time, max_nodes, max_procs, max_runtime,
                   max_memory, allow_overuse, max_queues, queues, queue, max_partitions, partitions, partition, note,
                   job_list, input_file)

    @classmethod
    def from_bsctrace(cls, input_file):
        """
        Creates a SWFtrace object from logs printed by the sim_mgr process of BSC's SLURM simulator
        :param input_file: location of the input file
        :return:
        """
        swf_version = "2.2"
        computer_name = ""
        location = ""
        acknowledge = ""
        conversion = ""
        max_jobs = ""
        max_records = ""
        preemption = ""
        unix_start_time = ""
        time_zone_string = ""
        max_nodes = ""
        max_procs = ""
        max_runtime = ""
        max_memory = ""
        allow_overuse = ""
        max_queues = ""
        queue = []
        max_partitions = ""
        partition = []
        note = "This is the result of {} ran on BSC's SLURM SIMULATOR"

        # read to pandas using fixed width format
        bscslurm_trace = pd.read_csv(input_file, sep=" ", header=None, names=["job_id", "uid", "gid", "username",
                                                                              "state", "partition", "time_limit",
                                                                              "submit_time", "start_time", "end_time",
                                                                              "node_list", "node_count", "proc_count",
                                                                              "work_dir", "backfilled"])
        # remove left part of the equal sign in the file
        bscslurm_trace = bscslurm_trace.apply(lambda x: x.str.replace(r"^[^=]+=", "").str.strip())

        # convert to numeric some fields
        bscslurm_trace[["job_id", "uid", "gid", "node_count", "proc_count", "submit_time", "start_time", "end_time"]] = \
            bscslurm_trace[
                ["job_id", "uid", "gid", "node_count", "proc_count", "submit_time", "start_time", "end_time"]].astype(
                int)

        # TODO: make the format of this dataframe the same as the class'

    @classmethod
    def from_cirne(cls, input_file):
        """
        Generates a swf object from the file generated by Cirne's Workload Model [3].

        :param input_file: location of the input file
        :return: SWFtrace file
        """
        swf_version = "2.2"
        computer_name = ""
        location = ""
        acknowledge = ""
        conversion = ""
        preemption = ""
        start_time = ""
        end_time = ""
        unix_start_time = ""
        time_zone_string = ""
        max_nodes = ""
        max_procs = ""
        max_runtime = ""
        max_memory = ""
        allow_overuse = ""
        queues = ""
        max_queues = ""
        queue = []
        partitions = ""
        max_partitions = ""
        partition = []
        note = "Trace generated from Walfredo Cirne's Model\n; Note: Cirne, Walfredo, and Francine Berman. 'A comprehensive model of the supercomputer workload'. IEEE, 2001."

        cirne_trace = pd.read_csv(input_file, sep="\s+", index_col=False, names=["job_name", "uid", "tag",
                                                                                     "job_type", "submit_time",
                                                                                     "alloc_cpus", "requested_time",
                                                                                     "start_time", "finish_time",
                                                                                     "status"])

        cirne_trace.loc[:, "wait_time"] = cirne_trace["start_time"] - cirne_trace["submit_time"]
        cirne_trace.loc[:, "runtime"] = cirne_trace["finish_time"] - cirne_trace["start_time"]
        cirne_trace.loc[cirne_trace["status"] == "completed", "status"] = 1
        cirne_trace.loc[cirne_trace["status"] == "cancelled", "status"] = 5
        max_jobs = str(len(cirne_trace))
        max_records = max_jobs
        job_list = cirne_trace[["submit_time", "alloc_cpus", "requested_time", "wait_time", "runtime", "status"]].copy()
        job_list.loc[:, "job_id"] = range(1, len(cirne_trace)+1)
        job_list.loc[:, "uid"] = -1
        job_list.loc[:, "gid"] = -1
        job_list.loc[:, "avg_cpu_used"] = -1
        job_list.loc[:, "used_memory"] = -1
        job_list.loc[:, "qos_id"] = -1
        job_list.loc[:, "partition_id"] = -1
        job_list.loc[:, "requested_memory"] = -1
        job_list.loc[:, "requested_cpus"] = -1
        job_list.loc[:, "app_id"] = -1
        job_list.loc[:, "preceding"] = -1
        job_list.loc[:, "thinktime"] = -1

        return cls(swf_version, computer_name, location, acknowledge, conversion, max_jobs, max_records, preemption,
                   unix_start_time, time_zone_string, start_time, end_time, max_nodes, max_procs, max_runtime,
                   max_memory, allow_overuse, max_queues, queues, queue, max_partitions, partitions, partition, note,
                   job_list, input_file)

    @property
    def swf_version(self):
        return self._swf_version

    @swf_version.setter
    def swf_version(self, version):
        self._swf_version = version

    @property
    def computer(self):
        return self._computer

    @computer.setter
    def computer(self, computer):
        self._computer = computer

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, location):
        self._location = location

    @property
    def acknowledge(self):
        return self._acknowledge

    @acknowledge.setter
    def acknowledge(self, acknowledge):
        self._acknowledge = acknowledge

    @property
    def conversion(self):
        return self._conversion

    @conversion.setter
    def conversion(self, conversion):
        self._conversion = conversion

    @property
    def max_jobs(self):
        return self._max_jobs

    @max_jobs.setter
    def max_jobs(self, max_jobs):
        self._max_jobs = max_jobs

    @property
    def max_records(self):
        return self._max_records

    @max_records.setter
    def max_records(self, max_records):
        self._max_records = max_records

    @property
    def preemption(self):
        return self._preemption

    @preemption.setter
    def preemption(self, preemption):
        self._preemption = preemption

    @property
    def unix_start_time(self):
        return self._unix_start_time

    @unix_start_time.setter
    def unix_start_time(self, unix_start_time):
        self._unix_start_time = unix_start_time

    @property
    def time_zone_string(self):
        return self._time_zone_string

    @time_zone_string.setter
    def time_zone_string(self, time_zone_string):
        self._time_zone_string = time_zone_string

    @property
    def start_time(self):
        return self._start_time

    @start_time.setter
    def start_time(self, start_time):
        self._start_time = strfdatelinux(start_time)

    @property
    def end_time(self):
        return self._end_time

    @end_time.setter
    def end_time(self, end_time):
        self._end_time = strfdatelinux(end_time)

    @property
    def max_nodes(self):
        return self._max_nodes

    @max_nodes.setter
    def max_nodes(self, max_nodes):
        self._max_nodes = max_nodes

    @property
    def max_procs(self):
        return self._max_procs

    @max_procs.setter
    def max_procs(self, max_procs):
        self._max_procs = max_procs

    @property
    def max_runtime(self):
        return self._max_runtime

    @max_runtime.setter
    def max_runtime(self, max_runtime):
        self._max_runtime = max_runtime

    @property
    def max_memory(self):
        return self._max_memory

    @max_memory.setter
    def max_memory(self, max_memory):
        self._max_memory = max_memory

    @property
    def allow_overuse(self):
        return self._allow_overuse

    @allow_overuse.setter
    def allow_overuse(self, allow_overuse):
        self._allow_overuse = allow_overuse

    @property
    def max_queues(self):
        return self._max_queues

    @max_queues.setter
    def max_queues(self, max_queues):
        self._max_queues = max_queues

    @property
    def queues(self):
        # TODO how to set multiple lines of comment here
        return self._queues

    @queues.setter
    def queues(self, queues):
        # TODO how to set multiple lines of comment here
        self._queues = queues

    @property
    def queue(self):
        return self._queue

    @queue.setter
    def queue(self, queue):
        self._queue = queue

    @property
    def max_partitions(self):
        return self._max_partitions

    @max_partitions.setter
    def max_partitions(self, max_partitions):
        self._max_partitions = max_partitions

    @property
    def partitions(self):
        return self._partition

    @partitions.setter
    def partitions(self, partition):
        self._partition = partition

    @property
    def note(self):
        return self._note

    @note.setter
    def note(self, note):
        self._note = note

    @property
    def input_file(self):
        return self._input_file

    @input_file.setter
    def input_file(self, input_file):
        self._input_file = input_file

    @property
    def job_list(self) -> pd.DataFrame:
        return self._job_list

    @job_list.setter
    def job_list(self, job_list):
        self._job_list = job_list

    def unique_users(self):
        """
        Returns the list with combo uid and gid
        :return: list with the different users and their group
        """
        users = []  # list of tuples
        unique_uids = pd.unique(self.job_list["uid"])
        for uid in unique_uids:
            # pick the first appearance of gid related to that uid. It should always be the same!
            gid = self.job_list[self.job_list["uid"] == uid].iloc[0]["gid"]
            uid_gid = uid, gid
            users.append(uid_gid)

        return users

    def add_job(self, submit_time, runtime, alloc_cpus, uid, requested_time, wait_time=-1, avg_cpu_used=-1,
                used_memory=-1, requested_cpus=-1, requested_memory=-1, status=-1, gid=-1, app_id=-1, qos_id=-1,
                partition_id=-1, preceding=-1, thinktime=-1):
        # this is to preserve the submitting order
        sub_before = self.job_list[self._job_list["submit_time"] < submit_time]
        sub_after = self.job_list[self._job_list["submit_time"] >= submit_time]
        # shift by 1 those that were submitted after than the job that it is being added
        sub_after.loc[:, "job_id"] += 1
        sub_after.loc[sub_after["preceding"] != -1, "preceding"] += 2  # match with simulator's id
        new_job_id = len(sub_before) + 1

        # 1 element data frame
        single_job = pd.DataFrame([[new_job_id, submit_time, wait_time, runtime, alloc_cpus, avg_cpu_used, used_memory,
                                    requested_cpus, requested_time, requested_memory, status, uid, gid, app_id, qos_id,
                                    partition_id, preceding, thinktime]],
                                  columns=["job_id", "submit_time", "wait_time", "runtime", "alloc_cpus",
                                           "avg_cpu_used", "used_memory", "requested_cpus", "requested_time",
                                           "requested_memory", "status", "uid", "gid", "app_id", "qos_id",
                                           "partition_id", "preceding", "thinktime"])

        print(f"Adding job to id {len(sub_before)}")
        self.job_list = pd.concat([sub_before, single_job, sub_after])

    # output to file!
    def to_bsc_swf(self, output_filename: str, nrecs: int = 2000) -> None:
        """
        Outputs a file with only jobs, without header, with commas for each of the job's attributes.

        Moreover, bscswf expects username, not uid, partition name, not id, and also qos name, not qos_id.

        Then we adapt users by adding a "u" in the begging of the string. E.g. uid 1001

        :param output_filename: name of the (bsc) swf file
        :return: None
        """
        job_list = self.job_list.loc[:nrecs].copy()  # local copy of the job list
        job_list["username"] = job_list["uid"].apply(lambda x: "u" + str(x))
        job_list["qos"] = job_list["qos_id"].apply(lambda x: "q" + str(x))
        job_list["partition"] = job_list["partition_id"].apply(lambda x: "p" + str(x))

        # we have to set it to 0, otherwise slurm will fail...
        job_list.loc[job_list["preceding"] == -1, "preceding"] = 0

        # TODO fix indexes for jobs with dependencies

        # output to bsc's format.
        job_list.to_csv(output_filename, columns=["job_id", "submit_time", "wait_time", "runtime", "alloc_cpus",
                                                  "avg_cpu_used", "used_memory", "requested_cpus", "requested_time",
                                                  "requested_memory", "status", "username", "gid", "app_id", "qos",
                                                  "partition", "preceding", "thinktime"], index=False, header=False)

    def to_swf(self, swf_filename) -> None:
        """
        Output to a .swf file
        :param swf_filename: output name of swf file
        :return:
        """
        # find out the size of each of the fields
        jobid_width = len(str(self.job_list["job_id"].max()))
        submit_width = len(str(self.job_list["submit_time"].max()))
        # absolute here just because in simulations the difference might be negative :/
        waittime_width = len(str(self.job_list["wait_time"].max()))
        if len(str(self.job_list["wait_time"].min())) > waittime_width:
            waittime_width = len(str(self.job_list["wait_time"].min()))
        elapsed_width = len(str(self.job_list["runtime"].max()))
        alloccpus_width = len(str(self.job_list["alloc_cpus"].max()))
        avecpu_width = len(str(self.job_list["avg_cpu_used"].max()))
        reqcpus_width = len(str(self.job_list["requested_cpus"].max()))
        timelimit_width = len(str(self.job_list["requested_time"].max()))
        reqmem_width = len(str(self.job_list["requested_memory"].max()))
        # We check if -1 is in the values of state to see the maximum width it can have
        if -1 in self.job_list["status"].values:
            state_width = 2
        else:
            state_width = 1
        uid_width = len(str(self.job_list["uid"].max()))
        gid_width = len(str(self.job_list["gid"].max()))
        qos_width = len(str(self.job_list["qos_id"].max()))
        partition_width = len(str(self.job_list["partition_id"].max()))
        preeceding_width = len(str(self.job_list["preceding"].max()))
        thinktime_width = len(str(self.job_list["thinktime"].max()))

        with open(swf_filename, "w") as swf_output:
            # HEADER
            swf_output.write(f"; Version: {self.swf_version}\n")
            swf_output.write(f"; Computer: {self.computer}\n")
            swf_output.write(f"; Installation: {self.location}\n")
            swf_output.write(f"; Acknowledge: {self.acknowledge}\n")
            swf_output.write(";\n")
            swf_output.write(f"; Conversion: Automatically generated by pyswf\n")
            swf_output.write(f"; Conversion: Manuel G. Marciani @ BSC-CNS Earth Sciences \n")
            swf_output.write(";\n")
            swf_output.write(f"; UnixStartTime: {self.unix_start_time}\n")
            swf_output.write(f"; TimeZoneString: {self.time_zone_string}\n")
            swf_output.write(f"; StartTime: {self.start_time}\n")
            swf_output.write(f"; EndTime: {self.end_time}\n")
            swf_output.write(";\n")
            swf_output.write(f"; MaxJobs: {len(self.job_list)}\n")
            swf_output.write(f"; MaxRecords: {len(self.job_list)}\n")
            # MACHINE PARAMETERS
            swf_output.write("; MaxNodes: \n")
            swf_output.write("; MaxProcs: \n")
            swf_output.write("; Preemption: \n")
            swf_output.write(";\n")
            # QUEUES
            swf_output.write("; Queues: ----------- \n")
            swf_output.write("; Queues:  id   name  \n")
            swf_output.write("; Queues: ----------- \n")
            for queue_id, queue_name in enumerate(self.queues):
                swf_output.write(f"; Queue: {queue_id}   {queue_name}\n")
            swf_output.write(f"; MaxQueues: {len(self.queues)}\n")
            swf_output.write(";\n")
            # PARTITIONS
            swf_output.write("; Partitions: ----------- \n")
            swf_output.write("; Partitions:  id   name  \n")
            swf_output.write("; Partitions: ----------- \n")
            for partitions_id, partitions_name in enumerate(self.partitions):
                swf_output.write(f"; Partition: {partitions_id}   {partitions_name}\n")
            swf_output.write(f"; MaxPartitions: {len(self.partitions)}\n")
            swf_output.write(f"; Note: {self.note}\n")
            swf_output.write(";\n")
            for job_id, job in self.job_list.iterrows():
                jobid_string = "{0:>{width}}".format(job['job_id'], width=jobid_width)
                submit_string = "{0:>{width}}".format(int(job['submit_time']), width=submit_width)
                waittime_string = "{0:>{width}}".format(job['wait_time'], width=waittime_width)
                elapsed_string = "{0:>{width}}".format(job['runtime'], width=elapsed_width)
                alloccpus_string = "{0:>{width}}".format(job['alloc_cpus'], width=alloccpus_width)
                avecpu_string = "{0:>{width}}".format(job['avg_cpu_used'], width=avecpu_width)
                reqcpus_string = "{0:>{width}}".format(job['requested_cpus'], width=reqcpus_width)
                timelimit_string = "{0:>{width}}".format(job['requested_time'], width=timelimit_width)
                reqmem_string = "{0:>{width}}".format(job['requested_memory'], width=reqmem_width)
                state_string = "{0:>{width}}".format(job['status'], width=state_width)
                uid_string = "{0:>{width}}".format(job['uid'], width=uid_width)
                gid_string = "{0:>{width}}".format(job['gid'], width=gid_width)
                qos_string = "{0:>{width}}".format(job['qos_id'], width=qos_width)
                partition_string = "{0:>{width}}".format(job['partition_id'], width=partition_width)
                preeceding_string = "{0:>{width}}".format(job['preceding'], width=preeceding_width)
                thinktime_string = "{0:>{width}}".format(job['thinktime'], width=thinktime_width)

                swf_output.write(
                    f"{jobid_string} {submit_string} {waittime_string} {elapsed_string} {alloccpus_string} "
                    f"{avecpu_string} -1 {reqcpus_string} {timelimit_string} {reqmem_string} {state_string} "
                    f"{uid_string} {gid_string} -1 {qos_string} {partition_string} {preeceding_string} {thinktime_string}\n")
